import threading
import time

import eel
from loguru import logger

from controllers.ControllerChromium import ControllerChromium
from models.StateADC import StateADC

IS_ROS = False
try:
    import rospy
    import actionlib
    import adc_msgs.msg
    from std_msgs.msg import String
    from nav_msgs.msg import Odometry
    IS_ROS = True
except Exception as e:
    pass


class ControllerADC:
    __instance = None

    def __init__(self):
        if ControllerADC.__instance is not None:
            raise Exception("Cannot have 2 instances of singleton ControllerADC")

        self.simulate_timer = None
        self.simulate_path_idx = 0
        self.simulate_pos = {
            "x": 0,
            "y": 0,
            "z": 0,
            "dir_x": 0,
            "dir_y": 0,
            "dir_z": 0,
            "dir_w": 0,
        }

        if IS_ROS:
            rospy.init_node("alarm_flight_ui")

            self.adc_takeoff_client = actionlib.SimpleActionClient('/fci/takeoff', adc_msgs.msg.ADCTakeoffAction)
            self.adc_land_client = actionlib.SimpleActionClient('/fci/land', adc_msgs.msg.ADCLandAction)
            self.adc_flytopoi_client = actionlib.SimpleActionClient('/fci/fly_to_poi', adc_msgs.msg.ADCFlyToPOIAction)
            self.adc_yaw_client = actionlib.SimpleActionClient('/fci/yaw', adc_msgs.msg.ADCYawAction)

            rospy.Subscriber("/rovio/odometry", Odometry, self._on_odometry)

    @staticmethod
    def instance():
        if ControllerADC.__instance is None:
            ControllerADC.__instance = ControllerADC()
        return ControllerADC.__instance

    def _on_simulate(self):
        if ControllerChromium.instance().project is not None:

            layer_drone_path = ControllerChromium.instance().project.map_layers.layer_drone_path
            x = layer_drone_path.x[self.simulate_path_idx]
            y = layer_drone_path.y[self.simulate_path_idx]
            x, y = ControllerChromium.instance().convert_x_y_to_original(x, y)

            self.simulate_pos["dir_x"] = x - self.simulate_pos["x"]
            self.simulate_pos["dir_y"] = y - self.simulate_pos["y"]

            self.simulate_pos["x"] = x
            self.simulate_pos["y"] = y

            self.simulate_path_idx += 1
            if len(layer_drone_path.x) <= self.simulate_path_idx:
                self.simulate_path_idx = 0

            ControllerChromium.instance().set_drone_position(
                self.simulate_pos["x"],
                self.simulate_pos["y"],
                self.simulate_pos["z"],
                self.simulate_pos["dir_x"],
                self.simulate_pos["dir_y"],
                self.simulate_pos["dir_z"],
                self.simulate_pos["dir_w"],
            )

            self.simulate_timer = threading.Timer(1.0, self._on_simulate)
            self.simulate_timer.start()


    def _on_odometry(self, odom_msg):
        ControllerChromium.instance().set_drone_position(
            odom_msg.pose.pose.position.x,
            odom_msg.pose.pose.position.y,
            odom_msg.pose.pose.position.z,
            odom_msg.pose.pose.orientation.x,
            odom_msg.pose.pose.orientation.y,
            odom_msg.pose.pose.orientation.z,
            odom_msg.pose.pose.orientation.w,
        )

    def _on_done(self, goal_status, result):
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'on_done: {goal_status} {result}'
        )

    def _on_fly_to_poi_done(self, goal_status, result):
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'on_fly_to_poi_done: {goal_status} {result}'
        )
        eel.finished_current_fly_to_poi()

    def _on_feedback(self, feedback):
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'feedback: {feedback}'
        )

    def fly_to_poi(self, x, y, z, w, name_poi="POI"):
        ControllerChromium.instance().set_state_adc(StateADC.FlyToPOI)
        ControllerChromium.instance().add_log_info(
            'UI',
            f'fly_to_poi: {x} {y} {name_poi}'
        )
        if IS_ROS:
            goal = adc_msgs.msg.ADCFlyToPOIGoal(
                String(name_poi),
                x,
                y,
                z,
                w
            )
            self.adc_flytopoi_client.send_goal(goal, self._on_fly_to_poi_done, self._on_fly_to_poi_active, self._on_feedback)
        else:
            time.sleep(1)
            self._on_fly_to_poi_done(None, None)

    def _on_fly_to_poi_active(self):
        ControllerChromium.instance().set_state_adc(StateADC.FlyToPOIActive)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'fly_to_poi_active'
        )

    def stop(self):
        ControllerChromium.instance().set_state_adc(StateADC.StopStop)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'stop'
        )

        if ControllerChromium.instance().state_adc == StateADC.FlyToPOI or \
                ControllerChromium.instance().state_adc == StateADC.FlyToPOIActive:
            ControllerChromium.instance().set_state_adc(StateADC.FlyToPOICancel)
            if IS_ROS:
                self.adc_flytopoi_client.cancel_goal()

        if ControllerChromium.instance().state_adc == StateADC.TakeOff or \
                ControllerChromium.instance().state_adc == StateADC.TakeOffActive:
            ControllerChromium.instance().set_state_adc(StateADC.TakeOffCancel)
            if IS_ROS:
                self.adc_takeoff_client.cancel_goal()

        if ControllerChromium.instance().state_adc == StateADC.Land or \
                ControllerChromium.instance().state_adc == StateADC.LandActive:
            ControllerChromium.instance().set_state_adc(StateADC.LandCancel)
            if IS_ROS:
                self.adc_land_client.cancel_goal()

        if ControllerChromium.instance().state_adc == StateADC.Yaw or \
                ControllerChromium.instance().state_adc == StateADC.YawActive:
            ControllerChromium.instance().set_state_adc(StateADC.YawCancel)
            if IS_ROS:
                self.adc_yaw_client.cancel_goal()

    def yaw(self, degrees):
        ControllerChromium.instance().set_state_adc(StateADC.Yaw)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'yaw: {degrees}'
        )
        if IS_ROS:
            goal = adc_msgs.msg.ADCYawGoal(degrees)
            self.adc_yaw_client.send_goal(goal, self._on_done, self._on_yaw_active, self._on_feedback)


    def _on_yaw_active(self):
        ControllerChromium.instance().set_state_adc(StateADC.YawActive)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'yaw_active'
        )

    def take_off(self):
        ControllerChromium.instance().set_state_adc(StateADC.TakeOff)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'take_off'
        )
        if IS_ROS:
            goal = adc_msgs.msg.ADCTakeoffGoal()
            self.adc_takeoff_client.send_goal(goal, self._on_done, self._on_take_off_active, self._on_feedback)
        else:
            ControllerChromium.instance().add_log_info(
                'UI',
                f'Simulating path following'
            )
            if self.simulate_timer is not None:
                self.simulate_timer.cancel()
            self._on_simulate()

    def _on_take_off_active(self):
        ControllerChromium.instance().set_state_adc(StateADC.TakeOffActive)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'take_off_active'
        )

    def land(self):
        ControllerChromium.instance().set_state_adc(StateADC.Land)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'land'
        )
        if IS_ROS:
            goal = adc_msgs.msg.ADCLandGoal()
            self.adc_land_client.send_goal(goal, self._on_done, self._on_land_active, self._on_feedback)

    def _on_land_active(self):
        ControllerChromium.instance().set_state_adc(StateADC.LandActive)
        ControllerChromium.instance().add_log_info(
            'ADC',
            f'land_active'
        )